@@ -12,21 +12,18 @@
 #
 ##############################################################################
 
+from collections import deque
 import socket
 import sys
 import threading
 import time
 
 from .buffers import ReadOnlyFileBasedBuffer
-from .compat import Empty, Queue, reraise, tobytes
+from .compat import reraise, tobytes
 from .utilities import (
-    Forwarded,
-    PROXY_HEADERS,
     build_http_date,
-    clear_untrusted_headers,
     logger,
     queue_logger,
-    undquote,
 )
 
 rename_headers = {  # or keep them without the HTTP_ prefix added
@@ -46,20 +43,20 @@ hop_by_hop = frozenset((
 ))
 
 
-class JustTesting(Exception):
-    pass
-
 class ThreadedTaskDispatcher(object):
     """A Task Dispatcher that creates a thread for each task.
     """
-    stop_count = 0 # Number of threads that will stop soon.
+    stop_count = 0  # Number of threads that will stop soon.
+    active_count = 0  # Number of currently active threads
     logger = logger
     queue_logger = queue_logger
 
     def __init__(self):
-        self.threads = {} # { thread number -> 1 }
-        self.queue = Queue()
-        self.thread_mgmt_lock = threading.Lock()
+        self.threads = set()
+        self.queue = deque()
+        self.lock = threading.Lock()
+        self.queue_cv = threading.Condition(self.lock)
+        self.thread_exit_cv = threading.Condition(self.lock)
 
     def start_new_thread(self, target, args):
         t = threading.Thread(target=target, name='waitress', args=args)
@@ -67,27 +64,31 @@ class ThreadedTaskDispatcher(object):
         t.start()
 
     def handler_thread(self, thread_no):
-        threads = self.threads
-        try:
-            while threads.get(thread_no):
-                task = self.queue.get()
-                if task is None:
-                    # Special value: kill this thread.
+        while True:
+            with self.lock:
+                while not self.queue and self.stop_count == 0:
+                    # Mark ourselves as idle before waiting to be
+                    # woken up, then we will once again be active
+                    self.active_count -= 1
+                    self.queue_cv.wait()
+                    self.active_count += 1
+
+                if self.stop_count > 0:
+                    self.active_count -= 1
+                    self.stop_count -= 1
+                    self.threads.discard(thread_no)
+                    self.thread_exit_cv.notify()
                     break
-                try:
-                    task.service()
-                except Exception as e:
-                    self.logger.exception(
-                        'Exception when servicing %r' % task)
-                    if isinstance(e, JustTesting):
-                        break
-        finally:
-            with self.thread_mgmt_lock:
-                self.stop_count -= 1
-                threads.pop(thread_no, None)
+
+                task = self.queue.popleft()
+            try:
+                task.service()
+            except BaseException:
+                self.logger.exception(
+                    'Exception when servicing %r', task)
 
     def set_thread_count(self, count):
-        with self.thread_mgmt_lock:
+        with self.lock:
             threads = self.threads
             thread_no = 0
             running = len(threads) - self.stop_count
@@ -95,54 +96,50 @@ class ThreadedTaskDispatcher(object):
                 # Start threads.
                 while thread_no in threads:
                     thread_no = thread_no + 1
-                threads[thread_no] = 1
+                threads.add(thread_no)
                 running += 1
                 self.start_new_thread(self.handler_thread, (thread_no,))
+                self.active_count += 1
                 thread_no = thread_no + 1
             if running > count:
                 # Stop threads.
-                to_stop = running - count
-                self.stop_count += to_stop
-                for n in range(to_stop):
-                    self.queue.put(None)
-                    running -= 1
+                self.stop_count += running - count
+                self.queue_cv.notify_all()
 
     def add_task(self, task):
-        queue_depth = self.queue.qsize()
-        if queue_depth > 0:
-            self.queue_logger.warning(
-                "Task queue depth is %d" %
-                queue_depth)
-        try:
-            task.defer()
-            self.queue.put(task)
-        except:
-            task.cancel()
-            raise
+        with self.lock:
+            self.queue.append(task)
+            self.queue_cv.notify()
+            queue_size = len(self.queue)
+            idle_threads = (
+                len(self.threads) - self.stop_count - self.active_count)
+            if queue_size > idle_threads:
+                self.queue_logger.warning(
+                    "Task queue depth is %d", queue_size - idle_threads)
 
     def shutdown(self, cancel_pending=True, timeout=5):
         self.set_thread_count(0)
         # Ensure the threads shut down.
         threads = self.threads
         expiration = time.time() + timeout
-        while threads:
-            if time.time() >= expiration:
-                self.logger.warning(
-                    "%d thread(s) still running" %
-                    len(threads))
-                break
-            time.sleep(0.1)
-        if cancel_pending:
-            # Cancel remaining tasks.
-            try:
+        with self.lock:
+            while threads:
+                if time.time() >= expiration:
+                    self.logger.warning(
+                        "%d thread(s) still running", len(threads))
+                    break
+                self.thread_exit_cv.wait(0.1)
+            if cancel_pending:
+                # Cancel remaining tasks.
                 queue = self.queue
-                while not queue.empty():
-                    task = queue.get()
-                    if task is not None:
-                        task.cancel()
-            except Empty: # pragma: no cover
-                pass
-            return True
+                if len(queue) > 0:
+                    self.logger.warning(
+                        "Canceling %d pending task(s)", len(queue))
+                while queue:
+                    task = queue.popleft()
+                    task.cancel()
+                self.queue_cv.notify_all()
+                return True
         return False
 
 class Task(object):
@@ -188,12 +185,6 @@ class Task(object):
                     self.status.startswith('304')
                     )
 
-    def cancel(self):
-        self.close_on_finish = True
-
-    def defer(self):
-        pass
-
     def build_response_header(self):
         version = self.version
         # Figure out whether the connection should be closed.
@@ -362,14 +353,9 @@ class ErrorTask(Task):
 
     def execute(self):
         e = self.request.error
-        body = '%s\r\n\r\n%s' % (e.reason, e.body)
-        tag = '\r\n\r\n(generated by waitress)'
-        body = body + tag
-        self.status = '%s %s' % (e.code, e.reason)
-        cl = len(body)
-        self.content_length = cl
-        self.response_headers.append(('Content-Length', str(cl)))
-        self.response_headers.append(('Content-Type', 'text/plain'))
+        status, headers, body = e.to_response()
+        self.status = status
+        self.response_headers.extend(headers)
         if self.version == '1.1':
             connection = self.request.headers.get('CONNECTION', '').lower()
             if connection == 'close':
@@ -379,15 +365,17 @@ class ErrorTask(Task):
             # HTTP 1.0
             self.response_headers.append(('Connection', 'close'))
         self.close_on_finish = True
+        self.content_length = len(body)
         self.write(tobytes(body))
 
+
 class WSGITask(Task):
     """A WSGI task produces a response from a WSGI application.
     """
     environ = None
 
     def execute(self):
-        env = self.get_environment()
+        environ = self.get_environment()
 
         def start_response(status, headers, exc_info=None):
             if self.complete and not exc_info:
@@ -449,27 +437,27 @@ class WSGITask(Task):
             return self.write
 
         # Call the application to handle the request and write a response
-        app_iter = self.channel.server.application(env, start_response)
-
-        if app_iter.__class__ is ReadOnlyFileBasedBuffer:
-            # NB: do not put this inside the below try: finally: which closes
-            # the app_iter; we need to defer closing the underlying file.  It's
-            # intention that we don't want to call ``close`` here if the
-            # app_iter is a ROFBB; the buffer (and therefore the file) will
-            # eventually be closed within channel.py's _flush_some or
-            # handle_close instead.
-            cl = self.content_length
-            size = app_iter.prepare(cl)
-            if size:
-                if cl != size:
-                    if cl is not None:
-                        self.remove_content_length_header()
-                    self.content_length = size
-                self.write(b'') # generate headers
-                self.channel.write_soon(app_iter)
-                return
+        app_iter = self.channel.server.application(environ, start_response)
 
+        can_close_app_iter = True
         try:
+            if app_iter.__class__ is ReadOnlyFileBasedBuffer:
+                cl = self.content_length
+                size = app_iter.prepare(cl)
+                if size:
+                    if cl != size:
+                        if cl is not None:
+                            self.remove_content_length_header()
+                        self.content_length = size
+                    self.write(b'') # generate headers
+                    # if the write_soon below succeeds then the channel will
+                    # take over closing the underlying file via the channel's
+                    # _flush_some or handle_close so we intentionally avoid
+                    # calling close in the finally block
+                    self.channel.write_soon(app_iter)
+                    can_close_app_iter = False
+                    return
+
             first_chunk_len = None
             for chunk in app_iter:
                 if first_chunk_len is None:
@@ -503,230 +491,9 @@ class WSGITask(Task):
                                 self.content_bytes_written, cl),
                         )
         finally:
-            if hasattr(app_iter, 'close'):
+            if can_close_app_iter and hasattr(app_iter, 'close'):
                 app_iter.close()
 
-    def parse_proxy_headers(
-        self,
-        environ,
-        headers,
-        trusted_proxy_count=1,
-        trusted_proxy_headers=None,
-    ):
-        if trusted_proxy_headers is None:
-            trusted_proxy_headers = set()
-
-        forwarded_for = []
-        forwarded_host = forwarded_proto = forwarded_port = forwarded = ""
-        client_addr = None
-        untrusted_headers = set(PROXY_HEADERS)
-
-        def warn_unspecified_behavior(header):
-            self.logger.warning(
-                "Found multiple values in %s, this has unspecified behaviour. "
-                "Ignoring header value.",
-                header,
-            )
-
-        if "x-forwarded-for" in trusted_proxy_headers and "X_FORWARDED_FOR" in headers:
-            forwarded_for = []
-
-            for forward_hop in headers["X_FORWARDED_FOR"].split(","):
-                forward_hop = forward_hop.strip()
-                forward_hop = undquote(forward_hop)
-
-                # Make sure that all IPv6 addresses are surrounded by brackets
-
-                if ":" in forward_hop and forward_hop[-1] != "]":
-                    forwarded_for.append("[{}]".format(forward_hop))
-                else:
-                    forwarded_for.append(forward_hop)
-
-            forwarded_for = forwarded_for[-trusted_proxy_count:]
-            client_addr = forwarded_for[0]
-
-            untrusted_headers.remove("X_FORWARDED_FOR")
-
-        if "x-forwarded-host" in trusted_proxy_headers and "X_FORWARDED_HOST" in headers:
-            forwarded_host_multiple = []
-
-            for forward_host in headers["X_FORWARDED_HOST"].split(","):
-                forward_host = forward_host.strip()
-                forward_host = undquote(forward_host)
-                forwarded_host_multiple.append(forward_host)
-
-            forwarded_host_multiple = forwarded_host_multiple[-trusted_proxy_count:]
-            forwarded_host = forwarded_host_multiple[0]
-
-            untrusted_headers.remove("X_FORWARDED_HOST")
-
-        if "x-forwarded-proto" in trusted_proxy_headers:
-            forwarded_proto = undquote(headers.get("X_FORWARDED_PROTO", ""))
-            untrusted_headers.remove("X_FORWARDED_PROTO")
-
-            if "," in forwarded_proto:
-                forwarded_proto = ""
-                warn_unspecified_behavior("X-Forwarded-Proto")
-
-        if "x-forwarded-port" in trusted_proxy_headers:
-            forwarded_port = undquote(headers.get("X_FORWARDED_PORT", ""))
-            untrusted_headers.remove("X_FORWARDED_PORT")
-
-            if "," in forwarded_port:
-                forwarded_port = ""
-                warn_unspecified_behavior("X-Forwarded-Port")
-
-        if "x-forwarded-by" in trusted_proxy_headers:
-            # Waitress itself does not use X-Forwarded-By, but we can not
-            # remove it so it can get set in the environ
-            untrusted_headers.remove("X_FORWARDED_BY")
-
-        if "forwarded" in trusted_proxy_headers:
-            forwarded = headers.get("FORWARDED", None)
-            untrusted_headers = PROXY_HEADERS - {"FORWARDED"}
-
-        # If the Forwarded header exists, it gets priority
-        if forwarded:
-            proxies = []
-
-            for forwarded_element in forwarded.split(","):
-                # Remove whitespace that may have been introduced when
-                # appending a new entry
-                forwarded_element = forwarded_element.strip()
-
-                forwarded_for = forwarded_host = forwarded_proto = ""
-                forwarded_port = forwarded_by = ""
-
-                for pair in forwarded_element.split(";"):
-                    pair = pair.lower()
-
-                    if not pair:
-                        continue
-
-                    token, equals, value = pair.partition("=")
-
-                    if equals != "=":
-                        raise ValueError("Invalid forwarded-pair in Forwarded element")
-
-                    if token.strip() != token:
-                        raise ValueError("token may not be surrounded by whitespace")
-
-                    if value.strip() != value:
-                        raise ValueError("value may not be surrounded by whitespace")
-
-                    if token == "by":
-                        forwarded_by = undquote(value)
-
-                    elif token == "for":
-                        forwarded_for = undquote(value)
-
-                    elif token == "host":
-                        forwarded_host = undquote(value)
-
-                    elif token == "proto":
-                        forwarded_proto = undquote(value)
-
-                    else:
-                        self.logger.warning("Unknown Forwarded token: %s" % token)
-
-                proxies.append(
-                    Forwarded(
-                        forwarded_by, forwarded_for, forwarded_host, forwarded_proto
-                    )
-                )
-
-            proxies = proxies[-trusted_proxy_count:]
-
-            # Iterate backwards and fill in some values, the oldest entry that
-            # contains the information we expect is the one we use. We expect
-            # that intermediate proxies may re-write the host header or proto,
-            # but the oldest entry is the one that contains the information the
-            # client expects when generating URL's
-            #
-            # Forwarded: for="[2001:db8::1]";host="example.com:8443";proto="https"
-            # Forwarded: for=192.0.2.1;host="example.internal:8080"
-            #
-            # (After HTTPS header folding) should mean that we use as values:
-            #
-            # Host: example.com
-            # Protocol: https
-            # Port: 8443
-
-            for proxy in proxies[::-1]:
-                client_addr = proxy.for_ or client_addr
-                forwarded_host = proxy.host or forwarded_host
-                forwarded_proto = proxy.proto or forwarded_proto
-
-        if forwarded_proto:
-            forwarded_proto = forwarded_proto.lower()
-
-            if forwarded_proto not in {"http", "https"}:
-                raise ValueError(
-                    'Invalid "Forwarded Proto=" or "X-Forwarded-Proto" value.'
-                )
-
-            # Set the URL scheme to the proxy provided proto
-            environ["wsgi.url_scheme"] = forwarded_proto
-
-            if not forwarded_port:
-                if forwarded_proto == "http":
-                    forwarded_port = "80"
-
-                if forwarded_proto == "https":
-                    forwarded_port = "443"
-
-        if forwarded_host:
-            if ":" in forwarded_host and forwarded_host[-1] != "]":
-                host, port = forwarded_host.rsplit(":", 1)
-                host, port = host.strip(), str(port)
-
-                # We trust the port in the Forwarded Host/X-Forwarded-Host over
-                # X-Forwarded-Port, or whatever we got from Forwarded
-                # Proto/X-Forwarded-Proto.
-
-                if forwarded_port != port:
-                    forwarded_port = port
-
-                # We trust the proxy server's forwarded Host
-                environ["SERVER_NAME"] = host
-                environ["HTTP_HOST"] = forwarded_host
-            else:
-                # We trust the proxy server's forwarded Host
-                environ["SERVER_NAME"] = forwarded_host
-                environ["HTTP_HOST"] = forwarded_host
-
-                if forwarded_port:
-                    if forwarded_port not in {"443", "80"}:
-                        environ["HTTP_HOST"] = "{}:{}".format(
-                            forwarded_host, forwarded_port
-                        )
-                    elif (
-                        forwarded_port == "80" and environ["wsgi.url_scheme"] != "http"
-                    ):
-                        environ["HTTP_HOST"] = "{}:{}".format(
-                            forwarded_host, forwarded_port
-                        )
-                    elif (
-                        forwarded_port == "443"
-                        and environ["wsgi.url_scheme"] != "https"
-                    ):
-                        environ["HTTP_HOST"] = "{}:{}".format(
-                            forwarded_host, forwarded_port
-                        )
-
-        if forwarded_port:
-            environ["SERVER_PORT"] = str(forwarded_port)
-
-        if client_addr:
-            if ":" in client_addr and client_addr[-1] != "]":
-                addr, port = client_addr.rsplit(":", 1)
-                environ["REMOTE_ADDR"] = addr.strip()
-                environ["REMOTE_PORT"] = port.strip()
-            else:
-                environ["REMOTE_ADDR"] = client_addr.strip()
-
-        return untrusted_headers
-
     def get_environment(self):
         """Returns a WSGI environment."""
         environ = self.environ
@@ -762,6 +529,14 @@ class WSGITask(Task):
                     path = path[len(url_prefix):]
 
         environ = {
+            'REMOTE_ADDR': channel.addr[0],
+            # Nah, we aren't actually going to look up the reverse DNS for
+            # REMOTE_ADDR, but we will happily set this environment variable
+            # for the WSGI application. Spec says we can just set this to
+            # REMOTE_ADDR, so we do.
+            'REMOTE_HOST': channel.addr[0],
+            # try and set the REMOTE_PORT to something useful, but maybe None
+            'REMOTE_PORT': str(channel.addr[1]),
             'REQUEST_METHOD': request.command.upper(),
             'SERVER_PORT': str(server.effective_port),
             'SERVER_NAME': server.server_name,
@@ -784,43 +559,12 @@ class WSGITask(Task):
             'wsgi.file_wrapper': ReadOnlyFileBasedBuffer,
             'wsgi.input_terminated': True,  # wsgi.input is EOF terminated
         }
-        remote_peer = environ['REMOTE_ADDR'] = channel.addr[0]
 
-        headers = dict(request.headers)
-
-        untrusted_headers = PROXY_HEADERS
-        if server.adj.trusted_proxy == '*' or remote_peer == server.adj.trusted_proxy:
-            untrusted_headers = self.parse_proxy_headers(
-                environ,
-                headers,
-                trusted_proxy_count=server.adj.trusted_proxy_count,
-                trusted_proxy_headers=server.adj.trusted_proxy_headers,
-            )
-        else:
-            # If we are not relying on a proxy, we still want to try and set
-            # the REMOTE_PORT to something useful, maybe None though.
-            environ["REMOTE_PORT"] = str(channel.addr[1])
-
-        # Nah, we aren't actually going to look up the reverse DNS for
-        # REMOTE_ADDR, but we will happily set this environment variable for
-        # the WSGI application. Spec says we can just set this to REMOTE_ADDR,
-        # so we do.
-        environ["REMOTE_HOST"] = environ["REMOTE_ADDR"]
-
-        # Clear out the untrusted proxy headers
-        if server.adj.clear_untrusted_proxy_headers:
-            clear_untrusted_headers(
-                headers,
-                untrusted_headers,
-                log_warning=server.adj.log_untrusted_proxy_headers,
-                logger=self.logger,
-            )
-
-        for key, value in headers.items():
+        for key, value in dict(request.headers).items():
             value = value.strip()
             mykey = rename_headers.get(key, None)
             if mykey is None:
-                mykey = 'HTTP_%s' % key
+                mykey = 'HTTP_' + key
             if mykey not in environ:
                 environ[mykey] = value
 
